from sklearn.naive_bayes import GaussianNB
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-gaussiannb')
class GaussianNBContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "var_smoothing": [
                0.000000001,
                0.000000002,
                0.000000005,
                0.000000008,
                0.000000009,
                0.0000001,
                0.0000002,
                0.0000003,
                0.0000005,
                0.0000007,
                0.0000009,
                0.00001,
                0.001,
                0.002,
                0.003,
                0.004,
                0.005,
                0.007,
                0.009,
                0.004,
                0.005,
                0.006,
                0.007,
                0.008,
                0.009,
                0.01,
                0.1,
                1,
            ]
        }
        tune_distributions = {
            "var_smoothing": UniformDistribution(0.000000001, 1, log=True)
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = GaussianNB(**kwargs)




from sklearn.gaussian_process import GaussianProcessClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-logisticregression')
class LogisticRegressionContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "max_iter_predict": [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000,]
        }
        tune_distributions = {"max_iter_predict": IntUniformDistribution(100, 1000)}
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = GaussianProcessClassifier(**kwargs)




from google.cloud import storage
from google.cloud import bigquery
import pickle
import os
import pandas as pd

from zueskit.logger.logger import LOGGER


class CloudService(object):
    def __init__(self, project=None):
        super(CloudService, self).__init__()
        self.project = project
        if project is not None:
            self.client = bigquery.Client(project=project)
            self.storage_client = storage.Client(project=project)
        else:
            self.client = bigquery.Client()
            self.storage_client = storage.Client()

    def read_gbq(self, query_string):
        df = self.client.query(query_string).result().to_dataframe(progress_bar_type='tqdm')
        return df

    def create_table(self, query_string, project, dataset, table):
        query_string = f"CREATE OR REPLACE TABLE `{project}.{dataset}.{table}` AS\n" + query_string
        print('query_string ', query_string)
        self.client.query(query_string).result()
        print(f'Done Create OR REPLACE `{project}.{dataset}.{table}`')
        return True

    def download_frombgtogcs(self, project, dataset_id, table_id, bucket_name, source_blob_name):
        for blob_name in self.list_blobs(bucket_name=bucket_name, prefix=source_blob_name):
            if '.' in blob_name:
                self.delete_blob(bucket_name=bucket_name, blob_name=blob_name)
        destination_uri = "gs://{}/{}/data-*.csv".format(bucket_name, source_blob_name)
        dataset_ref = bigquery.DatasetReference(project, dataset_id)
        table_ref = dataset_ref.table(table_id)
        extract_job = self.client.extract_table(
            table_ref,
            destination_uri,
            location="US",
        )  # API request
        extract_job.result()
        print("Exported {}:{}.{} to {}".format(self.project, dataset_id, table_id, destination_uri))
        return True

    def list_blobs(self, bucket_name, prefix=None):
        blobs = self.storage_client.list_blobs(bucket_name, prefix=prefix)
        results = []
        for blob in blobs:
            results.append(blob.name)
        return results

    def download_blob(self, bucket_name, source_blob_name, destination_file_name):
        """Downloads a blob from the bucket."""
        bucket = self.storage_client.bucket(bucket_name)
        blob = bucket.blob(source_blob_name)

        # check file exist
        check = False
        blob_lst = self.storage_client.list_blobs(bucket_name, prefix=source_blob_name)
        for blob in blob_lst:
            if source_blob_name == blob.name:
                check = True

        if check == True:
            blob.download_to_filename(destination_file_name)
            print("Blob {} downloaded to {}".format(source_blob_name, destination_file_name))
        else:
            print('Blob {} not downloaded to {}'.format(source_blob_name, destination_file_name))

    def delete_blob(self, bucket_name, blob_name):
        bucket = self.storage_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob.delete()
        print('Blob {} deleted.'.format(blob_name))

    def upload_blob(self, bucket_name, source_file_name, destination_blob_name):
        bucket = self.storage_client.bucket(bucket_name)
        blob = bucket.blob(destination_blob_name)
        blob.upload_from_filename(source_file_name)
        print('File {} uploaded to {}'.format(source_file_name, destination_blob_name))

    def download_object(self, bucket_name, gcs_file_name, local_file_name):
        self.download_blob(bucket_name, gcs_file_name, local_file_name)
        with open(local_file_name, 'rb') as fp:
            obj = pickle.load(fp)
        return obj

    def download_multipleobject(self, bucket_name, gcs_folder_name, local_folder_name):

        for blob_name in self.list_blobs(bucket_name=bucket_name, prefix=gcs_folder_name):
            if 'csv' in blob_name:
                self.download_blob(bucket_name, blob_name, os.path.join(local_folder_name, blob_name.split('/')[-1]))
        dfs = []
        for blob_name in os.listdir(local_folder_name):
            obj = pd.read_csv(os.path.join(local_folder_name, blob_name))
            dfs.append(obj)
        return pd.concat(dfs)

    def upload_object(self, object_name, bucket_name, local_file_name, gcs_file_name, format_file_name=None):
        try:
            if format_file_name == 'df':
                object_name.to_pickle(local_file_name)
            else:
                with open(local_file_name, 'wb') as fp:
                    pickle.dump(object_name, fp, protocol=pickle.HIGHEST_PROTOCOL)

            self.upload_blob(bucket_name, local_file_name, gcs_file_name)
        except:
            print('Fail upload object')
            return False
        return True
    def dump_object(self, object_name, local_file_name, format = None):
        try:
            if format == 'df':
                object_name.to_pickle(local_file_name)
            else:
                with open(local_file_name, 'wb') as fp:
                    pickle.dump(object_name, fp, protocol=pickle.HIGHEST_PROTOCOL)

        except:
            print('Fail dump object {}'.format(local_file_name))
            return False
        return True



import pickle

from sklearn.pipeline import Pipeline
from .logger import LOGGER



def save_model(model, model_path):
    LOGGER.info(f"Initializing save_model({model.__class__.__name__}, {model_path})")
    with open(model_path, 'wb') as fp:
        pickle.dump(model, fp, protocol=pickle.HIGHEST_PROTOCOL)
    LOGGER.info('Preprocess and Model Succesfully Saved')
    LOGGER.info('"save_model() succesfully completed......................................"')

def load_model(model_path):
    LOGGER.info(f"Initializing load_model({model_path})")
    with open(model_path, 'rb') as fp:
        obj = pickle.load(fp)
        LOGGER.info("Transformation Pipeline and Model Successfully Loaded")
    return obj

def supports_partial_fit(estimator, params: dict = None) -> bool:
    # special case for MLP
    from sklearn.neural_network import MLPClassifier

    if isinstance(estimator, MLPClassifier):
        try:
            if (
                params and "solver" in params and "lbfgs" in list(params["solver"])
            ) or estimator.solver == "lbfgs":
                return False
        except:
            return False

    if isinstance(estimator, Pipeline):
        return hasattr(estimator.steps[-1][1], "partial_fit")

    return hasattr(estimator, "partial_fit")
def can_early_stop(
    estimator, consider_partial_fit, consider_warm_start, consider_xgboost, params,
):
    """
    From https://github.com/ray-project/tune-sklearn/blob/master/tune_sklearn/tune_basesearch.py.

    Helper method to determine if it is possible to do early stopping.
    Only sklearn estimators with ``partial_fit`` or ``warm_start`` can be early
    stopped. warm_start works by picking up training from the previous
    call to ``fit``.

    Returns
    -------
        bool
            if the estimator can early stop
    """


    from sklearn.tree import BaseDecisionTree
    from sklearn.ensemble import BaseEnsemble

    try:
        base_estimator = estimator.steps[-1][1]
    except:
        base_estimator = estimator

    if consider_partial_fit:
        can_partial_fit = supports_partial_fit(base_estimator, params=params)
    else:
        can_partial_fit = False

    if consider_warm_start:
        is_not_tree_subclass = not issubclass(type(base_estimator), BaseDecisionTree)
        is_ensemble_subclass = issubclass(type(base_estimator), BaseEnsemble)
        can_warm_start = hasattr(base_estimator, "warm_start") and (
            (
                hasattr(base_estimator, "max_iter")
                and is_not_tree_subclass
                and not is_ensemble_subclass
            )
            or (is_ensemble_subclass and hasattr(base_estimator, "n_estimators"))
        )
    else:
        can_warm_start = False

    if consider_xgboost:
        from xgboost.sklearn import XGBModel

        is_xgboost = isinstance(base_estimator, XGBModel)
    else:
        is_xgboost = False

    LOGGER.info(
        f"can_partial_fit: {can_partial_fit}, can_warm_start: {can_warm_start}, is_xgboost: {is_xgboost}"
    )

    return can_partial_fit or can_warm_start or is_xgboost



from sklearn.ensemble import ExtraTreesClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution

@ModelFactory.register('classification-extratreesclassifier')
class ExtraTreesClassifierContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "n_estimators": np_list_arange(10, 300, 10, inclusive=True),
            "criterion": ["gini", "entropy"],
            "max_depth": np_list_arange(1, 11, 1, inclusive=True),
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "max_features": [1.0, "sqrt", "log2"],
            "bootstrap": [True, False],
            "min_samples_split": [2, 5, 7, 9, 10],
            "min_samples_leaf": [2, 3, 4, 5, 6],
            "class_weight": ["balanced", "balanced_subsample", {}],
        }
        tune_distributions = {
            "n_estimators": IntUniformDistribution(10, 300),
            "max_depth": IntUniformDistribution(1, 11),
            "min_samples_split": IntUniformDistribution(2, 10),
            "min_samples_leaf": IntUniformDistribution(1, 5),
            "max_features": UniformDistribution(0.4, 1),
            "min_impurity_decrease": UniformDistribution(0.000000001, 0.5, log=True),
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = ExtraTreesClassifier(**kwargs)




from sklearn.svm import SVC
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-svc')
class SVCClassifierContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "C": np_list_arange(0, 50, 0.01, inclusive=True),
            "class_weight": ["balanced", {}],
        }
        tune_distributions = {
            "C": UniformDistribution(0, 50),
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = SVC(**kwargs)




import numpy as np
from sklearn import metrics

class ClassificationMetricContainer():
    def __init__(self):
        super(ClassificationMetricContainer, self).__init__()
        self.report = {}

    def classification_report(self, y_true, y_pred):
        if 'accuracy_score' not in self.report.keys():
            self.report['accuracy_score'] = []
        self.report['accuracy_score'].append(metrics.accuracy_score(y_true, y_pred))

        if 'recall' not in self.report.keys():
            self.report['recall'] = []
        try:
            self.report['recall'].append(metrics.recall_score(y_true, y_pred))
        except:
            pass
        if 'precision' not in self.report.keys():
            self.report['precision'] = []
        try:
            self.report['precision'].append(metrics.precision_score(y_true, y_pred))
        except:
            pass
        if 'f1_score' not in self.report.keys():
            self.report['f1_score'] = []
        try: 
            self.report['f1_score'].append(metrics.f1_score(y_true, y_pred))
        except:
            pass
        if 'cohen_kappa' not in self.report.keys():
            self.report['cohen_kappa'] = []
        try:
            self.report['cohen_kappa'].append(metrics.cohen_kappa_score(y_true, y_pred))
        except:
            pass

        if 'mcc' not in self.report.keys():
            self.report['mcc'] = []
        try:
            self.report['mcc'].append(metrics.matthews_corrcoef(y_true, y_pred))
        except:
            pass
    def classification_report_proba(self, y_true, y_pred_proba):
        if 'auc' not in self.report.keys():
            self.report['auc'] = []
        try:
            self.report['auc'].append(metrics.roc_auc_score(y_true, y_pred_proba))
        except:
            pass
    def score_mean(self):
        for key, value in self.report.items():
            self.report[key] = np.mean(value)
        return self.report



from sklearn.ensemble import AdaBoostClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-adaboostclassifier')
class AdaBoostClassifierContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "n_estimators": np_list_arange(10, 300, 10, inclusive=True),
            "learning_rate": np_list_arange(0.001, 0.5, 0.001, inclusive=True),
            "algorithm": ["SAMME", "SAMME.R"],
        }
        tune_distributions = {
            "n_estimators": IntUniformDistribution(10, 300),
            "learning_rate": UniformDistribution(0.000001, 0.5, log=True),
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = AdaBoostClassifier(**kwargs)




from sklearn.ensemble import RandomForestClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils import IntUniformDistribution, UniformDistribution, np_list_arange
@ModelFactory.register('classification-randomforestclassifier')
class RandomForestClassifierContainer(RandomForestClassifier):
    def __init__(self, **kwargs):
        super(RandomForestClassifierContainer, self).__init__(**kwargs)
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "n_estimators": np_list_arange(10, 300, 10, inclusive=True),
            "max_depth": np_list_arange(1, 11, 1, inclusive=True),
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "max_features": [1.0, "sqrt", "log2"],
            "bootstrap": [True, False],
        }
        tune_distributions = {
            "n_estimators": IntUniformDistribution(10, 300),
            "max_depth": IntUniformDistribution(1, 11),
            "min_impurity_decrease": UniformDistribution(0.000000001, 0.5, log=True),
            "max_features": UniformDistribution(0.4, 1),
        }
        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = RandomForestClassifier(**kwargs)




from sklearn.neighbors import KNeighborsClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-kneighborsclassifier')
class KNeighborsClassifierContainer(KNeighborsClassifier):
    def __init__(self, **kwargs):
        super(KNeighborsClassifierContainer, self).__init__(**kwargs)
        tune_grid = {}
        tune_distributions = {}

        # common
        tune_grid["n_neighbors"] = range(1, 30)
        tune_grid["weights"] = ["uniform"]
        tune_grid["metric"] = ["minkowski", "euclidean", "manhattan"]


        tune_distributions["n_neighbors"] = IntUniformDistribution(1, 30)

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = KNeighborsClassifier(**kwargs)



import os
from enum import Enum, auto
import numpy as np
import pandas as pd
import json
from copy import deepcopy
import sklearn
from sklearn.datasets import make_classification
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score
from sklearn.ensemble import BaggingClassifier, AdaBoostClassifier, VotingClassifier, StackingClassifier

from autotonne.preprocessor import Preprocess
from autotonne.models.model_factory import ModelFactory
from autotonne.models.classification import *

#Tuning
import optuna
from ray.tune.sklearn import TuneGridSearchCV
from ray.tune.sklearn import TuneSearchCV
# visualization
from yellowbrick.classifier import ClassificationReport, ConfusionMatrix, ROCAUC, PrecisionRecallCurve, ClassPredictionError, DiscriminationThreshold
import matplotlib.pyplot as plt

from autotonne.utils.distributions import get_optuna_distributions

from autotonne.utils import LOGGER, can_early_stop
class Classification(object):
    def __init__(self,
                 X,
                 y,
                 preprocess: bool = True,
                 **kwargs
                 ):
        super(Classification, self).__init__()
        self.preprocess = preprocess
        print(kwargs)
        if self.preprocess == True:
            self.preprocessor = Preprocess(**kwargs)
        if self.preprocess == True:
            self.preprocessor.fit(X, y)
            X, y = self.preprocessor.transform(X, y)
        X = pd.DataFrame(X).reset_index(drop=True)
        y = pd.DataFrame(y).reset_index(drop=True)
        self.X = X
        self.y = y
        self.estimator = {}
        self.metrics = {}

    def create_model(self,
                      estimator,
                      cv: int  = 2,
                      scoring = ['roc_auc_ovr'],
                      sort = None,
                      fit_params = {},
                      n_jobs = -1,
                      verbose = False,
                      estimator_params = {}):
        """
        fit_kwargs: dict, default = {} (empty dict)
            Dictionary of arguments passed to the fit method of the model.
        **kwargs:
            Additional keyword arguments to pass to the estimator.
        """
        X = self.X
        y = self.y
        if sort is None:
            sort = scoring[0]
        score_models = {}
        estimators = {}
        if estimator is None:
            if len(self.estimator.keys()) > 0:
                for name_model, estimator in self.estimator.items():
                    estimators[name_model] = estimator
            else:
                for name_model in ModelFactory.name_registry:
                    if name_model in estimator_params.keys():
                        estimators[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                    else:
                        estimators[name_model] = ModelFactory.create_executor(name_model)
        else:
            for name_model in estimator:
                if name_model in estimator_params.keys():
                    estimators[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                else:
                    estimators[name_model] = ModelFactory.create_executor(name_model)
        for name_model, model in estimators.items():
            try:
                estimator = model.estimator
            except:
                estimator = model

            scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                                            X = X,
                                                            y = y,
                                                            scoring=scoring,
                                                            cv = cv,
                                                            n_jobs = n_jobs,
                                                            verbose = verbose,
                                                            fit_params = fit_params,
                                                            return_train_score = True,
                                                            return_estimator = True,
                                                            error_score=-1)
            self.estimator[name_model] = scores['estimator'][np.argmax(scores['test_'+sort])]
            scores.pop('estimator')
            name_model = ''.join(name_model.split('-')[1:])
            for key, values in scores.items():
                for i, value in enumerate(values):
                    if name_model not in self.metrics.keys():
                        self.metrics[name_model] = {}
                    self.metrics[name_model][key + "_{}fold".format(i + 1)] = value
        return self


    def compare_model(self,
                        cv: int  = 2,
                        scoring = ['roc_auc_ovr'],
                        sort = None,
                        fit_params = {},
                        n_jobs = -1,
                        verbose = False,
                        estimator_params = {}):
        X = self.X
        y = self.y
        if sort is None:
            sort = scoring[0]
        for name_model in ModelFactory.name_registry:
            if 'classification' in name_model:
                if name_model in estimator_params.keys():
                    LOGGER.info(f'Load estimator_params with {name_model}')
                    estimator_param = estimator_params[name_model]
                else:
                    estimator_param = estimator_params
                model = ModelFactory.create_executor(name_model, **estimator_param)
                estimator = model.estimator
                scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                                                X = X,
                                                                y = y,
                                                                scoring=scoring,
                                                                cv = cv,
                                                                n_jobs = n_jobs,
                                                                verbose = verbose,
                                                                fit_params = fit_params,
                                                                return_train_score = True,
                                                                return_estimator = True,
                                                                error_score=-1)
                self.estimator[name_model] = scores['estimator'][np.argmax(scores['test_'+sort])]
                scores.pop('estimator')
                name_model = ''.join(name_model.split('-')[1:])
                for key, values in scores.items():
                    for i, value in enumerate(values):
                        if name_model not in self.metrics.keys():
                            self.metrics[name_model] = {}
                        self.metrics[name_model][key + "_{}fold".format(i + 1)] = value

        return self


    def tune_model(self,
                    estimator = None,
                    n_iter = 2,
                    optimize = 'accuracy',
                    search_library: str = 'optuna',
                    search_algorithm = 'random',
                    early_stopping = 'asha',
                    early_stopping_max_iters = 10,
                    verbose = True,
                    n_jobs = -1
                    ):
        X = self.X
        y = self.y
        LOGGER.info('TUNE MODEL')
        best_params_model = {}
        model_grid = None

        estimator_tune = {}
        if estimator is None:
            for name_model in ModelFactory.name_registry:
                estimator_tune[name_model] = ModelFactory.create_executor(name_model)
        else:
            for name_model in estimator:
                estimator_tune[name_model] = ModelFactory.create_executor(name_model)

        for name_model, model in estimator_tune.items():
            LOGGER.info('tunning model_name: {}'.format(name_model))
            estimator = model.estimator

            parameter_grid = model.tune_grid
            parameter_distributions = model.tune_distributions
            if (search_library == 'scikit-learn' or search_library == 'tune-sklearn') and (search_algorithm == 'grid' or search_algorithm == 'random'):
                parameter_grid = model.tune_grid
            else:
                parameter_grid = model.tune_distributions
            model_grid = None
            if search_library == 'optuna':
                pruner_translator = {
                    "asha": optuna.pruners.SuccessiveHalvingPruner(),
                    "hyperband": optuna.pruners.HyperbandPruner(),
                    "median": optuna.pruners.MedianPruner(),
                    False: optuna.pruners.NopPruner(),
                    None: optuna.pruners.NopPruner(),
                }
                pruner = early_stopping
                if pruner in pruner_translator:
                    pruner = pruner_translator[early_stopping]

                sampler_translator = {
                    "tpe": optuna.samplers.TPESampler(seed=24),
                    "random": optuna.samplers.RandomSampler(seed=24),
                }
                sampler = sampler_translator[search_algorithm]

                try:
                    param_grid = get_optuna_distributions(parameter_distributions)
                except:
                    LOGGER.warn(
                        "Couldn't convert param_grid to specific library distributions. Exception:"
                    )
                    LOGGER.warn(traceback.format_exc())
                study = optuna.create_study(
                    direction = 'maximize', sampler = sampler, pruner = pruner
                )
                LOGGER.info('Initializing optuna.intergration.OptnaSearchCV')
                model_grid = optuna.integration.OptunaSearchCV(
                    estimator = estimator,
                    param_distributions = param_grid,
                    max_iter = early_stopping_max_iters,
                    n_jobs = n_jobs,
                    n_trials = n_iter,
                    random_state = 24,
                    scoring = optimize,
                    study = study,
                    refit = False,
                    verbose = verbose,
                    error_score = 'raise'
                )
            elif search_library == 'tune-sklearn':
                early_stopping_translator = {
                                            "asha": "ASHAScheduler",
                                            "hyperband": "HyperBandScheduler",
                                            "median": "MedianStoppingRule",
                                        }
                if early_stopping in early_stopping_translator:
                    early_stopping = early_stopping_translator[early_stopping]
                do_early_stop = early_stopping and can_early_stop(estimator, True, True, True, parameter_grid)

                if search_algorithm == 'grid':

                    LOGGER.info('Initializing tune_sklearn.TuneGridSearchCV')
                    model_grid = TuneGridSearchCV(
                        estimator = estimator,
                        param_grid = parameter_grid,
                        early_stopping = do_early_stop,
                        scoring = optimize,
                        cv = fold,
                        max_iters=early_stopping_max_iters,
                        refit = True,
                        n_jobs = n_jobs
                    )



            model_grid.fit(X, y)
            best_params = model_grid.best_params_
            LOGGER.info('best_params: {}'.format(best_params))
            best_params_model[name_model] = best_params
        return best_params_model
    def ensemble_model(self,
                       estimator = None,
                       method = 'bagging',
                       n_estimators = 2,
                       cv = 2,
                       scoring = ['roc_auc_ovr'],
                       sort = None,
                       fit_params = {},
                       verbose = True,
                       estimator_params = {},
                       n_jobs = -1):
        X = self.X
        y = self.y
        LOGGER.info('ensemble models')

        if sort is None:
            sort = scoring[0]
        score_models = {}
        estimator_ensemble = {}
        if estimator is None:
            if len(self.estimator.keys()) > 0:
                for name_model, estimator in self.estimator.items():
                    estimator_ensemble[name_model] = estimator
            else:
                for name_model in ModelFactory.name_registry:
                    if name_model in estimator_params.keys():
                        estimator_ensemble[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                    else:
                        estimator_ensemble[name_model] = ModelFactory.create_executor(name_model)
        else:
            for name_model in estimator:
                if name_model in estimator_params.keys():
                    estimator_ensemble[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                else:
                    estimator_ensemble[name_model] = ModelFactory.create_executor(name_model)

        for name_model, model in estimator_ensemble.items():
            try:
                estimator = model.estimator
            except:
                estimator = model
            if method == 'bagging':
                LOGGER.info('Initializing BaggingClassifer')
                estimator = BaggingClassifier(base_estimator=estimator,
                                                               n_estimators=n_estimators,
                                                               random_state=0)
            elif method == 'adaboost':
                LOGGER.info('Initializing AdaBoostClassifier')
                estimator = AdaBoostClassifier(base_estimator=estimator,
                                                                n_estimators=n_estimators,
                                                                random_state=0)

            scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                                            X = X,
                                                            y = y,
                                                            scoring=scoring,
                                                            cv = cv,
                                                            n_jobs = n_jobs,
                                                            verbose = verbose,
                                                            fit_params = fit_params,
                                                            return_train_score = True,
                                                            return_estimator = True,
                                                            error_score=-1)
            self.estimator[name_model] = scores['estimator'][np.argmax(scores['test_'+sort])]
            scores.pop('estimator')
            name_model = ''.join(name_model.split('-')[1:])
            for key, values in scores.items():
                for i, value in enumerate(values):
                    if name_model not in self.metrics.keys():
                        self.metrics[name_model] = {}
                    self.metrics[name_model][key + "_{}fold".format(i + 1)] = value
        return self
    def voting_model(self,
                     estimator = None,
                       cv = 2,
                       scoring = ['roc_auc_ovr'],
                       sort = None,
                       fit_params = {},
                       verbose = True,
                       estimator_params = {},
                       n_jobs = -1):
        X = self.X
        y = self.y
        LOGGER.info('VOTING MODELs')

        if sort is None:
            sort = scoring[0]
        score_models = {}
        estimator_voting = {}
        if estimator is None:
            if len(self.estimator.keys()) > 0:
                for name_model, estimator in self.estimator.items():
                    estimator_voting[name_model] = estimator
            else:
                for name_model in ModelFactory.name_registry:
                    if name_model in estimator_params.keys():
                        estimator_voting[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                    else:
                        estimator_voting[name_model] = ModelFactory.create_executor(name_model)
        else:
            for name_model in estimator:
                if name_model in estimator_params.keys():
                    estimator_voting[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                else:
                    estimator_voting[name_model] = ModelFactory.create_executor(name_model)
        model_voting = []
        for name_model, model in estimator_voting.items():
            try:
                estimator = model.estimator
            except:
                estimator = model
            model_voting.append((name_model, estimator))
        name_model = 'classification-votingclassifer'
        try:
            LOGGER.info('TRY soft voting')
            estimator = VotingClassifier(estimators= model_voting,
                                     voting='soft')
            scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                                X = X,
                                                y = y,
                                                scoring=scoring,
                                                cv = cv,
                                                n_jobs = n_jobs,
                                                verbose = verbose,
                                                fit_params = fit_params,
                                                return_train_score = True,
                                                return_estimator = True,
                                                error_score=-1)
        except:
            LOGGER.warn('TRY hard voting')
            estimator = VotingClassifier(estimators= model_voting,
                                     voting='hard')
            scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                                X = X,
                                                y = y,
                                                scoring=scoring,
                                                cv = cv,
                                                n_jobs = n_jobs,
                                                verbose = verbose,
                                                fit_params = fit_params,
                                                return_train_score = True,
                                                return_estimator = True,
                                                error_score=-1)
        self.estimator['classification-votingclassifer'] = scores['estimator'][np.argmax(scores['test_'+sort])]
        scores.pop('estimator')
        name_model = ''.join(name_model.split('-')[1:])
        for key, values in scores.items():
            for i, value in enumerate(values):
                if name_model not in self.metrics.keys():
                    self.metrics[name_model] = {}
                self.metrics[name_model][key + "_{}fold".format(i + 1)] = value
        return self

    def stacking_model(self,
                     estimator = None,
                    final_estimator = sklearn.linear_model.LogisticRegression(),
                       cv = 2,
                       scoring = ['roc_auc_ovr'],
                       sort = None,
                       fit_params = {},
                       verbose = True,
                       estimator_params = {},
                       n_jobs = -1):
        X = self.X
        y = self.y
        LOGGER.info('VOTING MODELs')

        if sort is None:
            sort = scoring[0]
        score_models = {}
        estimator_voting = {}
        if estimator is None:
            if len(self.estimator.keys()) > 0:
                for name_model, estimator in self.estimator.items():
                    estimator_voting[name_model] = estimator
            else:
                for name_model in ModelFactory.name_registry:
                    if name_model in estimator_params.keys():
                        estimator_voting[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                    else:
                        estimator_voting[name_model] = ModelFactory.create_executor(name_model)
        else:
            for name_model in estimator:
                if name_model in estimator_params.keys():
                    estimator_voting[name_model] = ModelFactory.create_executor(name_model, **estimator_params[name_model])
                else:
                    estimator_voting[name_model] = ModelFactory.create_executor(name_model)
        model_voting = []
        for name_model, model in estimator_voting.items():
            try:
                estimator = model.estimator
            except:
                estimator = model
            model_voting.append((name_model, estimator))
        name_model = 'classification-stacking_model'
        LOGGER.info('TRY STACKING MODEL')
        estimator = StackingClassifier(estimators= model_voting,
                                       final_estimator=final_estimator,
                                       cv = cv,
                                       n_jobs= n_jobs,
                                       verbose=verbose)
        scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                            X = X,
                                            y = y,
                                            scoring=scoring,
                                            cv = cv,
                                            n_jobs = n_jobs,
                                            verbose = verbose,
                                            fit_params = fit_params,
                                            return_train_score = True,
                                            return_estimator = True,
                                            error_score=-1)
        self.estimator['classification-stackingclassifer'] = scores['estimator'][np.argmax(scores['test_'+sort])]
        scores.pop('estimator')
        name_model = ''.join(name_model.split('-')[1:])
        for key, values in scores.items():
            for i, value in enumerate(values):
                if name_model not in self.metrics.keys():
                    self.metrics[name_model] = {}
                self.metrics[name_model][key + "_{}fold".format(i + 1)] = value
        return self
    def plot_model(self):
        LOGGER.info('Initializing plot model')
        size_plot = len(self.estimator.items())
        # fig, axes = plt.subplots(size_plot*6, figsize=(20, 20*size_plot*6))
        if os.path.isdir(os.path.join(os.getcwd(), 'viz')) == False:
            os.makedirs(os.path.join(os.getcwd(), 'viz/'))
        classes = pd.value_counts(self.y.values.flatten()).index
        index = 0
        for idx, (name_model, estimator) in enumerate(self.estimator.items()):
            X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(self.X, self.y, test_size = 0.2, stratify=self.y, random_state = 24)
            try:
                viz = ClassificationReport(model = estimator)
                index +=1
                viz.fit(X_train, y_train)
                viz.score(X_test, y_test)
                viz.show(outpath=os.path.join(os.getcwd(), f'viz/{viz.__class__.__name__}_{estimator.__class__.__name__}_{index}.png'), clear_figure=True)
                index = index + 1
            except:
                pass
            try:
                viz = ConfusionMatrix(model = estimator)
                viz.fit(X_train, y_train)
                viz.score(X_test, y_test)
                viz.show(outpath=os.path.join(os.getcwd(), f'viz/{viz.__class__.__name__}_{estimator.__class__.__name__}_{index}.png'), clear_figure=True)
                index = index + 1
            except:
                pass
            try:
                viz = ROCAUC(model = estimator)
                viz.fit(X_train, y_train)
                viz.score(X_test, y_test)
                viz.show(outpath=os.path.join(os.getcwd(), f'viz/{viz.__class__.__name__}_{estimator.__class__.__name__}_{index}.png'), clear_figure=True)
                index = index + 1
            except:
                pass
            try:
                viz = PrecisionRecallCurve(model = estimator, per_class=True)
                viz.fit(X_train, y_train)
                viz.score(X_test, y_test)
                viz.show(outpath=os.path.join(os.getcwd(), f'viz/{viz.__class__.__name__}_{estimator.__class__.__name__}_{index}.png'), clear_figure=True)
                index = index + 1
            except:
                pass
            try:
                viz = ClassPredictionError(model = estimator, classes = classes)
                viz.fit(X_train, y_train)
                viz.score(X_test, y_test)
                viz.show(outpath=os.path.join(os.getcwd(), f'viz/{viz.__class__.__name__}_{estimator.__class__.__name__}_{index}.png'), clear_figure=True)
                index = index + 1
            except:
                LOGGER.warn(f'{viz.__class__.__name__} ERROR')
            try:
                viz = DiscriminationThreshold(model = estimator)
                viz.fit(X_train, y_train)
                viz.score(X_test, y_test)
                viz.show(outpath=os.path.join(os.getcwd(), f'viz/{viz.__class__.__name__}_{estimator.__class__.__name__}_{index}.png'), clear_figure=True)
                index = index + 1
            except:
                LOGGER.warn(f'{viz.__class__.__name__} ERROR')
    def report_classification(self, sort_by=None):
        scores = pd.DataFrame.from_dict(self.metrics, orient = 'index')
        if sort_by is not None:
            scores = scores.sort_values(by = sort_by, ascending=False)
        return scores







from sklearn.linear_model import RidgeClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution

@ModelFactory.register('classification-ridgeclassifier')
class RidgeClassifierContainer(RidgeClassifier):
    def __init__(self, **kwargs):
        # super(RidgeClassifierContainer, self).__init__()
        super().__init__()
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "normalize": [True, False],
        }

        tune_grid["alpha"] = np_list_arange(0.01, 10, 0.01, inclusive=False)
        tune_grid["fit_intercept"] = [True, False]
        tune_distributions["alpha"] = UniformDistribution(0.001, 10)

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = RidgeClassification(**kwargs)

class RidgeClassification(RidgeClassifier):
    def predict_proba(self, X):
        return self.predict(X)



from sklearn.neural_network import MLPRegressor
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution

@ModelFactory.register('regression-mlpregressor')
class MLPRegressorContainer(MLPRegressor):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        tune_grid = {
            "learning_rate": ["constant", "invscaling", "adaptive"],
            "alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                0.9,
            ],
            "hidden_layer_size_0": [50, 100],
            "hidden_layer_size_1": [0, 50, 100],
            "hidden_layer_size_2": [0, 50, 100],
            "activation": ["tanh", "identity", "logistic", "relu"],
        }
        tune_distributions = {
            "alpha": UniformDistribution(0.0000000001, 0.9999999999, log=True),
            "hidden_layer_size_0": IntUniformDistribution(50, 100),
            "hidden_layer_size_1": IntUniformDistribution(0, 100),
            "hidden_layer_size_2": IntUniformDistribution(0, 100),
        }

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = MLPRegressor(**kwargs)



from sklearn.linear_model import ElasticNet
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution

@ModelFactory.register('regression-elasticnet')
class ElasticNetRegressionContainer(ElasticNet):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        tune_grid = {
            "alpha": np_list_arange(0.01, 10, 0.01, inclusive=True),
            "l1_ratio": np_list_arange(0.01, 1, 0.001, inclusive=False),
            "fit_intercept": [True, False],
            "normalize": [True, False],
        }
        tune_distributions = {
            "alpha": UniformDistribution(0, 1),
            "l1_ratio": UniformDistribution(0.01, 0.9999999999),
        }

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = ElasticNet(**kwargs)



from xgboost import XGBRegressor
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution

@ModelFactory.register('regression-xgbregressor')
class XGBRegressorContainer(XGBRegressor):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        tune_grid = {
            "learning_rate": np_list_arange(0.001, 0.5, 0.001, inclusive=True),
            "n_estimators": np_list_arange(10, 300, 10, inclusive=True),
            "subsample": [0.2, 0.3, 0.5, 0.7, 0.9, 1],
            "max_depth": np_list_arange(1, 11, 1, inclusive=True),
            "colsample_bytree": [0.5, 0.7, 0.9, 1],
            "min_child_weight": [1, 2, 3, 4],
            "reg_alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "reg_lambda": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "scale_pos_weight": np_list_arange(0, 50, 0.1, inclusive=True),
        }
        tune_distributions = {
            "learning_rate": UniformDistribution(0.000001, 0.5, log=True),
            "n_estimators": IntUniformDistribution(10, 300),
            "subsample": UniformDistribution(0.2, 1),
            "max_depth": IntUniformDistribution(1, 11),
            "colsample_bytree": UniformDistribution(0.5, 1),
            "min_child_weight": IntUniformDistribution(1, 4),
            "reg_alpha": UniformDistribution(0.0000000001, 10, log=True),
            "reg_lambda": UniformDistribution(0.0000000001, 10, log=True),
            "scale_pos_weight": UniformDistribution(1, 50),
        }

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = XGBRegressor(**kwargs)



from .catboost_regressor import CatBoostRegressorContainer
from .elasticnet import ElasticNetRegressionContainer
from .gradientboosting_regressor import GradientBoostingRegressorContainer
from .lars import LarsRegressionContainer
from .lasso_regression import LassoRegressionContainer
from .lgbm_regressor import LGBMRegressorContainer
from .mlp_regressor import MLPRegressorContainer
from .ridge_regression import RidgeRegressionContainer



from sklearn.linear_model import Lars
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution

@ModelFactory.register('regression-lars')
class LarsRegressionContainer(Lars):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        tune_grid = {
            "fit_intercept": [True, False],
            "normalize": [True, False],
            "eps": [
                0.00001,
                0.0001,
                0.001,
                0.01,
                0.05,
                0.0005,
                0.005,
                0.00005,
                0.02,
                0.007,
                0.1,
            ],
        }
        tune_distributions = {
            "eps": UniformDistribution(0.00001, 0.1),
        }

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = Lars(**kwargs)



import logging

def get_logger() -> logging.Logger:
    try:
        assert bool(LOGGER)
        return LOGGER
    except:
        return create_logger()


def create_logger() -> logging.Logger:
    logger = logging.getLogger("logs")
    logger.setLevel(logging.DEBUG)

    # create console handler and set level to debug
    if logger.hasHandlers():
        logger.handlers.clear()

    ch = logging.FileHandler("logs.log")
    ch.setLevel(logging.DEBUG)

    # create formatter
    formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)

    return logger


LOGGER = create_logger()



from catboost import CatBoostRegressor
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution

@ModelFactory.register('regression-catboostregressor')
class CatBoostRegressorContainer(CatBoostRegressor):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        tune_grid = {
            "depth": list(range(1, 12)),
            "n_estimators": np_list_arange(10, 300, 10, inclusive=True),
            "random_strength": np_list_arange(0, 0.8, 0.1, inclusive=True),
            "l2_leaf_reg": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 20, 30, 50, 100, 200],
        }
        tune_distributions = {
            "depth": IntUniformDistribution(1, 11),
            "n_estimators": IntUniformDistribution(10, 300),
            "random_strength": UniformDistribution(0, 0.8),
            "l2_leaf_reg": IntUniformDistribution(1, 200, log=True),
        }
        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = CatBoostRegressor(**kwargs)



import numpy as np
import pandas as  pd
from autotonne.base.model_base import ModelBase
from autotonne.base.object_factory import ObjectFactory
# models
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn.naive_bayes import GaussianNB

from autotonne.utils.distributions import *


@ObjectFactory.register('logistic_regression_container')
class LogisticRegressionContainer(LogisticRegression):
    def __init__(self, **kwargs):
        super(LogisticRegressionContainer, self).__init__(**kwargs)
        tune_args = {}
        tune_grid = {}
        tune_distributions = {}

        tune_grid['penalty'] = ['l2', 'none']
        tune_grid['C'] = np_list_arange(0, 10, 0.001, inclusive=True)
        tune_distributions['C'] = UniformDistribution(0, 10)



class KNeighborsClassifierContainer(KNeighborsClassifier):
    def __init__(self, **kwargs):
        super(KNeighborsClassifierContainer, self).__init__(**kwargs)
        tune_args = {}
        tune_grid = {}
        tune_distributions = {}

        tune_grid['penalty'] = ['l2', 'none']
        tune_grid['C'] = np_list_arange(0, 10, 0.001, inclusive=True)
        tune_distributions['C'] = UniformDistribution(0, 10)

# class GaussianNBContainer(GaussianNB):
#     def __init__(self, **kwargs)



from sklearn.tree import DecisionTreeClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-decisiontreeclassifier')
class DecisionTreeClassifierContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "max_depth": np_list_arange(1, 16, 1, inclusive=True),
            "max_features": [1.0, "sqrt", "log2"],
            "min_samples_leaf": [2, 3, 4, 5, 6],
            "min_samples_split": [2, 5, 7, 9, 10],
            "criterion": ["gini", "entropy"],
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
        }
        tune_distributions = {
            "max_depth": IntUniformDistribution(1, 16),
            "max_features": UniformDistribution(0.4, 1),
            "min_samples_leaf": IntUniformDistribution(2, 6),
            "min_samples_split": IntUniformDistribution(2, 10),
            "min_impurity_decrease": UniformDistribution(0.000000001, 0.5, log=True),
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = DecisionTreeClassifier(**kwargs)




from sklearn.cluster import KMeans
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution

@ModelFactory.register('clustering-kmeans')
class KMeansContainer(KMeans):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = KMeans(**kwargs)




from enum import Enum, auto
import numpy as np
import pandas as pd
import json
from copy import deepcopy
import sklearn
from sklearn.datasets import make_classification
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score

from autotonne.preprocessor import Preprocess
from autotonne.models.model_factory import ModelFactory
from autotonne.models.clustering import *

#Tuning
import optuna
from ray.tune.sklearn import TuneGridSearchCV
from ray.tune.sklearn import TuneSearchCV
from autotonne.utils.distributions import get_optuna_distributions

from autotonne.utils import LOGGER, can_early_stop
class Clustering(object):
    def __init__(self,
                 preprocess: bool = True,
                 **kwargs
                 ):
        super(Clustering, self).__init__()
        self.preprocess = preprocess
        if self.preprocess == True:
            self.preprocessor = Preprocess(**kwargs)
        self.estimator = {}

    def create_model(self,
                      X,
                      num_clusters,
                      estimator,
                      fit_params = {},
                      n_jobs = -1,
                      verbose = False,
                      estimator_params = {}):
        """
        fit_kwargs: dict, default = {} (empty dict)
            Dictionary of arguments passed to the fit method of the model.
        **kwargs:
            Additional keyword arguments to pass to the estimator.
        """
        if self.preprocess == True:
            self.preprocessor.fit(X, None)
            X = self.preprocessor.transform(X, None)
        X = pd.DataFrame(X).reset_index(drop=True)

        score_models = {}
        for name_model in ModelFactory.name_registry:
            if str(estimator) in name_model:
                if name_model in estimator_params.keys():
                    estimator_param = estimator_params[name_model]
                else:
                    estimator_param = estimator_params
                model = ModelFactory.create_executor(name_model, **estimator_param)
                estimator = model.estimator
                estimator.fit(X)
                self.estimator[name_model] = estimator
    def predict(self, X, sample_weight = None):
        if self.preprocess == True:
            X = self.preprocessor.transform(X)
        labels = {}
        for name_model, estimator in self.estimator.items():
            labels[name_model] = estimator.predict(X)
        return labels







from enum import Enum, auto
import numpy as np
import pandas as pd
import sklearn
from imblearn.pipeline import Pipeline
from autotonne.preprocess.preprocess import *
from autotonne.preprocess.preprocee_factory import PreprocessFactory
from autotonne.utils import LOGGER

class Preprocess(object):
    def __init__(self,
                datatype = True,
                datatype_categorical_columns = [],
                datatype_numeric_columns = [],
                datatype_time_columns = [],
                imputer = True,
                imputer_numeric_strategy = "mean",
                imputer_categorical_strategy = "most_frequent",

                zeronearzerovariance = True,
                zeronearzerovariance_threshold_first = 0.1,
                zeronearzerovariance_threshold_second = 20,

                categoryencoder = True,
                categoryencoder_cols = [],
                categoryencoder_method = 'targetencoder',

                groupsimilarfeature = True,
                groupsimilarfeature_group_name = ['OMG', 'ONEID'],
                groupsimilarfeature_list_of_group_feature = [['GKDiving', 'GKHandling'], ['GKKicking', 'GKReflexes']],

                binning = False,
                binning_features_to_discretize= [],

                maketimefeature =True,
                makefeature_time_columns = [],
                maketimefeature_list_of_feature = ['month',  'dayofweek', 'weekday', 'is_month_end', 'is_month_start', 'hour'],

                scaling = True,
                scaling_method = 'zscore',
                scaling_numeric_columns = [],

                outlier = True,
                outlier_method = ['pca', 'iforest', 'knn'],
                outlier_contamination = 0.2,

                makenonlinearfeature = True,
                makenonlinearfeature_polynomial_columns = [],
                makenonlinearfeature_degree = 2,
                makenonlinearfeature_intergration_only = False,
                makenonlinearfeature_include_bias = False,
                makenonlinearfeature_other_nonlinear_feature = ["sin", "cos", "tan"],

                rfe = False,
                rfe_estimator = None,
                rfe_step = 1,
                rfe_min_features_to_select = 3,
                rfe_cv = 3,

                reducedimension = False,
                reducedimension_method = "pca_linear",
                reducedimension_n_components = 0.99,
                random_state = 42,
                n_jobs = -1,
                 ):
        super(Preprocess, self).__init__()
        self.params = {
            'datatype': {
                'flag': datatype,
                'numeric_columns': datatype_numeric_columns,
                'categorical_columns': datatype_categorical_columns,
                'time_columns': datatype_time_columns
            },
            'simpleimputer': {
                'flag': imputer,
                'numeric_strategy': imputer_numeric_strategy,
                'categorical_strategy': imputer_categorical_strategy
            },

            'zeronearzerovariance': {
                'flag': zeronearzerovariance,
                'threshold_first': zeronearzerovariance_threshold_first,
                'threshold_second': zeronearzerovariance_threshold_second
            },

            'categoryencoder': {
                'flag': categoryencoder,
                'cols': categoryencoder_cols,
                'method': categoryencoder_method
            },

            'groupsimilarfeature': {
                'flag': groupsimilarfeature,
                'group_name': groupsimilarfeature_group_name,
                'list_of_group_feature': groupsimilarfeature_list_of_group_feature
            },

            'binning': {
                'flag': binning,
                'features_to_discretize': binning_features_to_discretize
            },
            'scaling': {
                'flag': scaling,
                'method': scaling_method,
                'numeric_columns': scaling_numeric_columns
            },

            'maketimefeature': {
                'flag': maketimefeature,
                'time_columns': makefeature_time_columns,
                'list_of_feature': maketimefeature_list_of_feature
            },
            'outlier': {
                'flag': outlier,
                'methods': outlier_method,
                'contamination': outlier_contamination
            },

            'makenonlinearfeature': {
                'flag': makenonlinearfeature,
                "polynomial_columns": makenonlinearfeature_polynomial_columns,
                "degree": makenonlinearfeature_degree,
                "interaction_only": makenonlinearfeature_intergration_only,
                "include_bias": makenonlinearfeature_include_bias,
                "other_nonlinear_feature": makenonlinearfeature_other_nonlinear_feature
            },
            'rfe': {
                'flag': rfe,
                'estimator': rfe_estimator,
                'step': rfe_step,
                'min_features_to_select': rfe_min_features_to_select,
                'cv': rfe_cv
            },
            'reducedimension': {
                'flag': reducedimension,
                'method': reducedimension_method,
                'n_components': reducedimension_n_components
            }
        }
        self.preprocess_objs = []
        for name_registry, preprocess_param in self.params.items():
            if preprocess_param['flag'] == False or preprocess_param['flag'] is None:
                continue
            preprocess_param.pop('flag')
            name_registry = '-'.join(['preprocess', name_registry])
            LOGGER.info('preprocess registry {}'.format(name_registry))
            preprocess_object = PreprocessFactory.create_executor(name = name_registry,
                                                                  **preprocess_param)
            self.preprocess_objs.append(preprocess_object)

    def fit(self, X, y = None, **fit_params):
        X = X.copy()
        if y is not None:
            y = y.copy()
            if isinstance(y, pd.Series):
                y = pd.DataFrame(y)
            y.columns = [str(col) for col in y.columns.tolist()]
        if isinstance(X, pd.Series):
            X = pd.DataFrame(X)
        X.columns = [str(col) for col in X.columns.tolist()]
        X = X.loc[:, ~X.columns.duplicated()]
        X.dropna(axis=1, how='all', inplace=True)
        for obj in self.preprocess_objs:
            obj.fit(X, y)

    def transform(self, X, y = None, **fit_params):
        X = X.copy()
        if y is not None:
            y = y.copy()
            if isinstance(y, pd.Series):
                y = pd.DataFrame(y)
            y.columns = [str(col) for col in y.columns.tolist()]

        if isinstance(X, pd.Series):
            X = pd.DataFrame(X)
        X.columns = [str(col) for col in X.columns.tolist()]
        X = X.loc[:, ~X.columns.duplicated()]
        X.dropna(axis=1, how='all', inplace=True)
        data = X, y
        for obj in self.preprocess_objs:
            data = obj.transform(*data)
        if data[1] is None:
            return data[0]
        return data
    def fit_transform(self, X, y = None):
        self.fit(X, y)
        return self.transform(X, y)

if __name__=='__main__':
    X, y = sklearn.datasets.load_linnerud(return_X_y=True, as_frame=True)
    data = Preprocess().fit_transform(X)
    print(data)






from sklearn.linear_model import LogisticRegression
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-logisticregression')
class LogisticRegressionContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid['penalty'] = ["l2", "none"]
        tune_grid['C'] = np_list_arange(0, 10, 1.0, inclusive=True)

        tune_distributions['C'] = UniformDistribution(0, 10)
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = LogisticRegression(**kwargs)




from enum import Enum, auto
import numpy as np
import pandas as pd
import json
from copy import deepcopy
import sklearn
from sklearn.datasets import make_classification
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score

from autotonne.preprocessor import Preprocess
from autotonne.models.model_factory import ModelFactory
from autotonne.models.regression import *

#Tuning
import optuna
from ray.tune.sklearn import TuneGridSearchCV
from ray.tune.sklearn import TuneSearchCV
from autotonne.utils.distributions import get_optuna_distributions

from autotonne.utils import LOGGER, can_early_stop
class Regression(object):
    def __init__(self,
                 preprocess: bool = True,
                 **kwargs
                 ):
        super(Regression, self).__init__()
        self.preprocess = preprocess
        print(kwargs)
        if self.preprocess == True:
            self.preprocessor = Preprocess(**kwargs)
        self.estimator = {}

    def create_model(self,
                      X,
                      y,
                      estimator,
                      cv: int  = 2,
                      scoring = ['max_error'],
                      sort = None,
                      fit_params = {},
                      n_jobs = -1,
                      verbose = False,
                      estimator_params = {}):
        """
        fit_kwargs: dict, default = {} (empty dict)
            Dictionary of arguments passed to the fit method of the model.
        **kwargs:
            Additional keyword arguments to pass to the estimator.
        """
        if self.preprocess == True:
            self.preprocessor.fit(X, y)
            X, y = self.preprocessor.transform(X, y)
        X = pd.DataFrame(X).reset_index(drop=True)
        y = pd.DataFrame(y).reset_index(drop=True)

        if sort is None:
            sort = scoring[0]
        score_models = {}
        for name_model in ModelFactory.name_registry:
            if str(estimator) in name_model:
                if name_model in estimator_params.keys():
                    estimator_param = estimator_params[name_model]
                else:
                    estimator_param = estimator_params
                model = ModelFactory.create_executor(name_model, **estimator_param)
                estimator = model.estimator
                scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                                                X = X,
                                                                y = y,
                                                                scoring=scoring,
                                                                cv = cv,
                                                                n_jobs = n_jobs,
                                                                verbose = verbose,
                                                                fit_params = fit_params,
                                                                return_train_score = True,
                                                                return_estimator = True)
                self.estimator[name_model] = scores['estimator'][np.argmax(scores['test_'+sort])]
                scores.pop('estimator')
                name_model = ''.join(name_model.split('-')[1:])
                for key, values in scores.items():
                    for i, value in enumerate(values):
                        if name_model not in score_models.keys():
                            score_models[name_model] = {}
                        score_models[name_model][key + "_{}fold".format(i + 1)] = value

        score_models = pd.read_json(json.dumps(score_models))
        print('score_models: ', score_models)
        return score_models


    def compare_model(self,
                        X,
                        y,
                        cv: int  = 2,
                        scoring = ['max_error'],
                        sort = None,
                        fit_params = {},
                        n_jobs = -1,
                        verbose = False,
                        estimator_params = {}):
        if self.preprocess == True:
            self.preprocessor.fit(X, y)
            X, y = self.preprocessor.transform(X, y)
        X = pd.DataFrame(X).reset_index(drop=True)
        y = pd.DataFrame(y).reset_index(drop=True)
        if sort is None:
            sort = scoring[0]
        score_models = {}
        for name_model in ModelFactory.name_registry:
            if 'regression' in name_model:
                if name_model in estimator_params.keys():
                    estimator_param = estimator_params[name_model]
                else:
                    estimator_param = estimator_params
                model = ModelFactory.create_executor(name_model, **estimator_param)
                estimator = model.estimator
                scores = sklearn.model_selection.cross_validate(estimator = estimator,
                                                                X = X,
                                                                y = y,
                                                                scoring=scoring,
                                                                cv = cv,
                                                                n_jobs = n_jobs,
                                                                verbose = verbose,
                                                                fit_params = fit_params,
                                                                return_train_score = True,
                                                                return_estimator = True)
                self.estimator[name_model] = scores['estimator'][np.argmax(scores['test_'+sort])]
                scores.pop('estimator')
                name_model = ''.join(name_model.split('-')[1:])
                for key, values in scores.items():
                    for i, value in enumerate(values):
                        if name_model not in score_models.keys():
                            score_models[name_model] = {}
                        score_models[name_model][key + "_{}fold".format(i + 1)] = value

        score_models = pd.read_json(json.dumps(score_models))
        print('score_models: ', score_models)
        return score_models



    def tune_model(self,
                    X,
                    y,
                    estimator = None,
                    fold = None,
                    n_iter = 2,
                    optimize = 'max_error',
                    search_library: str = 'optuna',
                    search_algorithm = 'random',
                    early_stopping = 'asha',
                    early_stopping_max_iters = 10,
                    verbose = True,
                    n_jobs = -1
                    ):
        if self.preprocess == True:
            self.preprocessor.fit(X, y)
            X, y = self.preprocessor.transform(X, y)
        X = pd.DataFrame(X).reset_index(drop=True)
        y = pd.DataFrame(y).reset_index(drop=True)
        LOGGER.info('tune models')
        best_params_model = {}
        model_grid = None

        estimator_tune = {}
        if estimator is None:
            for name_model in ModelFactory.name_registry:
                estimator_tune[name_model] = ModelFactory.create_executor(name_model)
        else:
            for name_model in estimator:
                estimator_tune[name_model] = ModelFactory.create_executor(name_model)

        for name_model, model in estimator_tune.items():
            LOGGER.info('tunning model_name: {}'.format(name_model))
            estimator = model.estimator

            parameter_grid = model.tune_grid
            parameter_distributions = model.tune_distributions
            if (search_library == 'scikit-learn' or search_library == 'tune-sklearn') and (search_algorithm == 'grid' or search_algorithm == 'random'):
                parameter_grid = model.tune_grid
            else:
                parameter_grid = model.tune_distributions
            model_grid = None
            if search_library == 'optuna':
                pruner_translator = {
                    "asha": optuna.pruners.SuccessiveHalvingPruner(),
                    "hyperband": optuna.pruners.HyperbandPruner(),
                    "median": optuna.pruners.MedianPruner(),
                    False: optuna.pruners.NopPruner(),
                    None: optuna.pruners.NopPruner(),
                }
                pruner = early_stopping
                if pruner in pruner_translator:
                    pruner = pruner_translator[early_stopping]

                sampler_translator = {
                    "tpe": optuna.samplers.TPESampler(seed=24),
                    "random": optuna.samplers.RandomSampler(seed=24),
                }
                sampler = sampler_translator[search_algorithm]

                try:
                    param_grid = get_optuna_distributions(parameter_distributions)
                except:
                    logger.warning(
                        "Couldn't convert param_grid to specific library distributions. Exception:"
                    )
                    logger.warning(traceback.format_exc())
                study = optuna.create_study(
                    direction = 'maximize', sampler = sampler, pruner = pruner
                )
                LOGGER.info('Initializing optuna.intergration.OptnaSearchCV')
                model_grid = optuna.integration.OptunaSearchCV(
                    estimator = estimator,
                    param_distributions = param_grid,
                    max_iter = early_stopping_max_iters,
                    n_jobs = n_jobs,
                    n_trials = n_iter,
                    random_state = 24,
                    scoring = optimize,
                    study = study,
                    refit = False,
                    verbose = verbose,
                    error_score = 'raise'
                )
            elif search_library == 'tune-sklearn':
                early_stopping_translator = {
                                            "asha": "ASHAScheduler",
                                            "hyperband": "HyperBandScheduler",
                                            "median": "MedianStoppingRule",
                                        }
                if early_stopping in early_stopping_translator:
                    early_stopping = early_stopping_translator[early_stopping]
                do_early_stop = early_stopping and can_early_stop(estimator, True, True, True, parameter_grid)

                if search_algorithm == 'grid':

                    LOGGER.info('Initializing tune_sklearn.TuneGridSearchCV')
                    model_grid = TuneGridSearchCV(
                        estimator = estimator,
                        param_grid = parameter_grid,
                        early_stopping = do_early_stop,
                        scoring = optimize,
                        cv = fold,
                        max_iters=early_stopping_max_iters,
                        refit = True,
                        n_jobs = n_jobs
                    )



            model_grid.fit(X, y)
            best_params = model_grid.best_params_
            LOGGER.info('best_params: {}'.format(best_params))
            best_params_model[name_model] = best_params
        return best_params_model





from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-quadraticdiscriminantanalysis')
class QuadraticDiscriminantAnalysisContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {"reg_param": np_list_arange(0, 1, 0.01, inclusive=True)}
        tune_distributions = {"reg_param": UniformDistribution(0, 1)}
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = QuadraticDiscriminantAnalysis(**kwargs)




from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution, CategoricalDistribution
@ModelFactory.register('classification-lineardiscriminantanalysis')
class LinearDiscriminantAnalysisContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "solver": ["lsqr", "eigen"],
            "shrinkage": [
                None,
                "auto",
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
                0.6,
                0.7,
                0.8,
                0.9,
                1,
            ],
        }
        tune_distributions = {
            "solver": CategoricalDistribution(['lsqr', 'eigen']),
            "shrinkage": UniformDistribution(0.0001, 1, log=True),
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = LinearDiscriminantAnalysis(**kwargs)




from sklearn.linear_model import SGDClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution
@ModelFactory.register('classification-sgdclassifier')
class SGDClassifierContainer():
    def __init__(self, tol = 0.001, loss = 'hinge', penalty = 'l2', eta0=0.001, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "penalty": ["elasticnet", "l2", "l1"],
            "l1_ratio": np_list_arange(0.0000000001, 1, 0.01, inclusive=False),
            "alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "fit_intercept": [True, False],
            "learning_rate": ["constant", "invscaling", "adaptive", "optimal"],
            "eta0": [0.001, 0.01, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5],
        }
        tune_distributions = {
            "l1_ratio": UniformDistribution(0.0000000001, 0.9999999999),
            "alpha": UniformDistribution(0.0000000001, 0.9999999999, log=True),
            "eta0": UniformDistribution(0.001, 0.5, log=True),
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = SGDClassifier(tol = tol, loss = loss, penalty=penalty, eta0=eta0, **kwargs)




import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from scipy import stats
import sklearn
import sklearn.impute
import sklearn.feature_selection
import pyod
import pyod.models.knn
import pyod.models.iforest
import pyod.models.pca
import category_encoders as ce

from autotonne.preprocess.preprocee_factory import PreprocessFactory
from autotonne.utils.logger import LOGGER

@PreprocessFactory.register('preprocess-datatype')
class DataTypes(BaseEstimator, TransformerMixin):
    def __init__(self,
                 categorical_columns = [],
                 numeric_columns = [],
                 time_columns = [],
                 verbose = True):
        self.categorical_columns = categorical_columns
        self.numeric_columns = numeric_columns
        self.time_columns = time_columns
        self.verbose = verbose
    @staticmethod
    def str_if_not_null(x):
        if pd.isnull(x) or (x is None) or pd.isna(x) or (x is not x):
            return x
        return str(x)
    def fit(self, X, y = None):
        LOGGER.info('FIT DATATYPE')
        X = X.copy()
        if y is not None:
            y = y.copy()
        X.replace([np.inf, -np.inf], np.NaN, inplace=True)
        X.columns = [str(col) for col in X.columns]
        for col in X.select_dtypes(include=['object']).columns:
            try:
                X[col] = X[col].astype("int64")
            except:
                None
        for col in X.select_dtypes(include=['object']).columns:
            try:
                X[col] = pd.to_datatime(X[col], infer_datetime_format = True, utc = False, errors = 'raise')
            except:
                continue
        # if data type is bool or pandas Categorical, convert to categorical
        for col in X.select_dtypes(include=["bool", "category"]).columns:
            X[col] = X[col].astype("object")

        # with csv, if we have any null in a column that was int, pandas will read it as float
        # so first we need to convert any such floats that hae NaN and unique values are lower than 20
        for col in X.select_dtypes(include = ["float64"]).columns:
            X[col] = X[col].astype("float32")

            nan_count = sum(X[col].isnull())
            count_float = np.nansum([False if r.is_integer() else True for r in X[col]])
            count_float = (count_float - nan_count)
            if (count_float == 0) and (X[col].nunique() <= 20) and (nan_count > 0):
                X[col] = X[col].astype("object")

            # if column is int and unique counts are more than two
            for col in X.select_dtypes(include = ["int64"]).columns:
                if X[col].nunique() <= 20:
                    X[col] = X[col].apply(self.str_if_not_null)
                else:
                    X[col] = X[col].astype("float32")
            for col in X.select_dtypes(include=['float32']).columns:
                if X[col].nunique() == 2:
                    X[col] = X[col].apply(self.str_if_not_null)

            for col in self.numeric_columns:
                try:
                    X[col] = X[col].astype("float32")
                except:
                    X[col] = X[col].astype(self.str_if_not_null)

            for col in self.categorical_columns:
                try:
                    X[col] = X[col].apply(self.str_if_not_null)
                except:
                    continue

            for col in self.time_columns:
                try:
                    X[col] = pd.to_datetime(X[col], infer_datetime_format=True, utc = False, errors = "raise")
                except:
                    continue
            for col in X.select_dtypes(include=["datetime64"]).columns:
                X[col] = X[col].astype("datetime64[ns]")
            self.learned_dtypes = X.dtypes

            X = X.replace([np.inf, -np.inf], np.NaN).astype(self.learned_dtypes)
            return self

    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM DATATYPE')
        X = X.copy()
        if y is not None:
            y = y.copy()
        X.replace([np.inf, -np.inf], np.nan, inplace=True)
        X.columns = [str(col) for col in X.columns]
        X = X.astype(self.learned_dtypes)

        # remove columns with duplicate name
        # X = X.loc[:, ~X.columns.duplicated()]

        # remove nas
        X.dropna(axis = 0, how='all', inplace = True)
        # X.dropna(axis = 1, how = 'all', inplace = True)

        return X, y
    def fit_transform(self, X, y = None):
        self.fit(X, y)
        return self.transform(X, y)




@PreprocessFactory.register('preprocess-simpleimputer')
class SimpleImputer(BaseEstimator, TransformerMixin):
    def __init__(self, numeric_strategy,
                 categorical_strategy,
                 fill_value_numerical = 0,
                 fill_value_categorical = 'not_available'):
        self.numeric_imputer = sklearn.impute.SimpleImputer(strategy=numeric_strategy, fill_value=fill_value_numerical)
        self.categorical_imputer = sklearn.impute.SimpleImputer(strategy=categorical_strategy, fill_value=fill_value_categorical)

    def fit(self, X, y = None):
        LOGGER.info('FIT SIMPLE IMPUTER')
        X = X.copy()
        if y is not None:
            y = y.copy()
        self.numeric_columns = [str(col) for col in X.select_dtypes(include=[np.number]).columns.tolist()]
        self.categorical_columns = [str(col) for col in X.select_dtypes(include=['object', 'category']).columns.tolist()]
        self.time_columns = [str(col) for col in X.select_dtypes(include=['datetime64[ns]']).columns.tolist()]
        if len(self.numeric_columns) > 0:
            self.numeric_imputer.fit(X.loc[:, self.numeric_columns])

        if len(self.categorical_columns) > 0:
            self.categorical_imputer.fit(X[self.categorical_columns])

        if len(self.time_columns) > 0:
            self.most_frequent_time = []
            for col in self.time_columns:
                self.most_frequent_time.append(X[col].mode()[0])

        return self
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM SIMPLE IMPUTER')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if len(self.numeric_columns) > 0:
            X.loc[:, self.numeric_columns] = self.numeric_imputer.transform(X.loc[:, self.numeric_columns])
        if len(self.categorical_columns) > 0:
            X.loc[:, self.categorical_columns] = self.categorical_imputer.transform(X.loc[:, self.categorical_columns])
        if len(self.time_columns) > 0:
            for idx, col in enumerate(self.time_columns):
                X[col].fillna(self.most_frequent_time[idx])
        return X, y
    def fit_transform(self, X, y = None):
        X = X.copy()
        self.fit(X, y)
        return self.transform(X, y)


# Zero and Near Zero Variance
@PreprocessFactory.register('preprocess-zeronearzerovariance')
class ZeroNearZeroVariance(BaseEstimator, TransformerMixin):
    def __init__(self, threshold_first = 0.1, threshold_second = 20):
        self.threshold_first = threshold_first
        self.threshold_second = threshold_second
        self.to_drop = []
    def fit(self, X, y = None):
        LOGGER.info('FIT ZERO AND NEARZERO Variance')
        X = X.copy()
        if y is not None:
            y = y.copy()

        for col in X.columns:
            u = pd.DataFrame(X[col].value_counts()).sort_values(by=col, ascending = False, inplace = False)
            first = len(u) / len(X)
            if len(u[col]) == 1:
                second = 100
            else:
                second = u.iloc[0, 0] / u.iloc[1, 0]

            if first <= self.threshold_first and second >= self.threshold_second:
                self.to_drop.append(col)
            if second == 100:
                self.to_drop.append(col)
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM ZERO and NEARZERO Variance')
        X = X.copy()
        if y is not None:
            y = y.copy()
        X = X.drop(columns = self.to_drop)
        return X, y
    def fit_transform(self, X, y = None):
        X = X.copy()
        if y is not None:
            y = y.copy()
        self.fit(X, y)
        return self.transform(X, y)


@PreprocessFactory.register('preprocess-categoryencoder')
class CategoryEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, cols = [], method='onehotencoder'):
        self.cols = cols
        self.method = method
    def fit(self, X, y = None):
        LOGGER.info('FIT Category-Encoder')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if len(self.cols) == 0:
            self.cols = X.select_dtypes(include=['category', 'object']).columns.tolist()
        self.encoder = self.choose_method(cols = self.cols, method = self.method)
        self.encoder.fit(X, y)
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM Category-Encoder')
        X = X.copy()
        if y is not None:
            y = y.copy()
        X = self.encoder.transform(X)
        return X, y
    def fit_transform(self, X, y = None):
        X = X.copy()
        self.fit(X, y)
        return self.transform(X, y)
    def choose_method(self, cols, method):
        if method == 'backwarddifferenceencoder':
            encoder = ce.BackwardDifferenceEncoder(cols=cols, return_df=True)
        elif method == 'baseencoder':
            encoder = ce.BaseNEncoder(cols=cols, return_df=True)
        elif method == 'binaryencoder':
            encoder = ce.BinaryEncoder(cols=cols, return_df=True)
        elif method == 'catboostencoder':
            encoder = ce.CatBoostEncoder(cols=cols, return_df=True)
        elif method == 'countencoder':
            encoder = ce.CountEncoder(cols=cols, return_df=True)
        elif method == 'glmmeencoder':
            encoder = ce.GLMMEncoder(cols=cols, return_df=True)
        elif method == 'hashingencoder':
            encoder = ce.HashingEncoder(cols=cols, return_df=True)
        elif method == 'helmerencoder':
            encoder = ce.HelmertEncoder(cols=cols, return_df=True)
        elif method == 'jamessteinencoder':
            encoder = ce.JamesSteinEncoder(cols=cols, return_df=True)
        elif method == 'leaveoneoutencoder':
            encoder = ce.LeaveOneOutEncoder(cols=cols, return_df=True)
        elif method == 'mestimateencoder':
            encoder = ce.MEstimateEncoder(cols=cols, return_df=True)
        elif method == 'onehotencoder':
            encoder = ce.OneHotEncoder(cols=cols, use_cat_names=True, return_df=True)
        elif method == 'ordinalencoder':
            encoder = ce.OrdinalEncoder(cols=cols, return_df=True)
        elif method == 'sumencoder':
            encoder = ce.SumEncoder(cols=cols, return_df=True)
        elif method == 'polynomialencoder':
            encoder = ce.PolynomialEncoder(cols=cols, return_df=True)
        elif method == 'targetencoder':
            encoder = ce.TargetEncoder(cols=cols, return_df=True)
        elif method == 'woeeencoder':
            encoder = ce.WOEEncoder(cols=cols, return_df=True)
        return encoder

@PreprocessFactory.register('preprocess-groupsimilarfeature')
class GroupSimilarFeature(BaseEstimator, TransformerMixin):
    def __init__(
            self,
            group_name =[],
            list_of_group_feature = []):
        self.list_of_group_feature = list_of_group_feature
        self.group_name = group_name
    def fit(self, X, y = None):
        LOGGER.info('FIT GroupSimilarFeature')
        X = X.copy()
        if y is not None:
            y = y.copy()
        return self
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM GroupSimilarFeature')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if len(self.list_of_group_feature) > 0:
            for f, g in zip(self.list_of_group_feature, self.group_name):
                X[g+"_min"] = X[f].apply(np.min, 1)
                X[g+"_max"] = X[f].apply(np.max, 1)
                X[g+"_mean"] = X[f].apply(np.mean, 1)
                X[g+"_median"] = X[f].apply(np.median, 1)
                X[g+"_mode"] = stats.mode(X[f], 1)[0]
                X[g+"_std"] = X[f].apply(np.std, 1)
        return X, y
    def fit_transform(self, X, y = None):
        X = X.copy()
        if y is not None:
            y = y.copy()
        self.fit(X, y)
        return self.transform(X, y)



@PreprocessFactory.register('preprocess-binning')
class Binning(BaseEstimator, TransformerMixin):
    def __init__(self, features_to_discretize):
        self.features_to_discretize = features_to_discretize
        self.binns = []
    def fit(self, X, y = None):
        LOGGER.info('FIT BINNING')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if len(self.features_to_discretize) > 0:
            for col in self.features_to_discretize:
                hist, bin_edg = np.histogram(X[col], bins = 'sturges')
                self.binns.append(len(hist))
            self.disc = sklearn.preprocessing.KBinsDiscretizer(n_bins=self.binns,
                                                               encode = 'ordinal',
                                                               strategy='kmeans').fit(X[self.features_to_discretize])
        return self
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM BINNING')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if len(self.features_to_discretize) > 0:
            X[self.features_to_discretize] = self.disc.transform(X[self.features_to_discretize])

        return X, y

@PreprocessFactory.register('preprocess-scaling')
class Scaling(BaseEstimator, TransformerMixin):
    def __init__(self, numeric_columns = 'not_available', method = 'zscore'):
        self.method = method
        self.numeric_columns = numeric_columns
    def fit(self, X, y = None):
        LOGGER.info('FIT SCALING')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if isinstance(self.numeric_columns, str) and self.numeric_columns == 'not_available':
            return self
        if len(self.numeric_columns) == 0:
            self.numeric_columns = X.select_dtypes(include=[np.number]).columns.tolist()
        if self.method == 'zscore':
            self.scale = sklearn.preprocessing.StandardScaler()
        elif self.method == 'minmax':
            self.scale = sklearn.preprocessing.MinMaxScaler()
        elif self.method == 'yj':
            self.scale = sklearn.preprocessing.PowerTransformer(method = 'yeo-johnson', standardize=True)
        elif self.method == 'quantile':
            self.scale = sklearn.preprocessing.QuantileTransformer(output_distribution= 'normal')
        elif self.method == 'robust':
            self.scale = sklearn.preprocessing.RobustScaler()
        elif self.method =='maxabs':
            self.scale = sklearn.preprocessing.MaxAbsScaler()
        self.scale.fit(X.loc[:, self.numeric_columns])
        return self
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM SCALING')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if isinstance(self.numeric_columns, str) and self.numeric_columns == 'not_available':
            return X
        X.loc[:, self.numeric_columns] = self.scale.transform(X.loc[:, self.numeric_columns])
        return X, y
    def fit_transform(self, X, y = None):
        self.fit(X, y)
        return self.transform(X, y)

@PreprocessFactory.register('preprocess-maketimefeature')
class MakeTimeFeature(BaseEstimator, TransformerMixin):
    def __init__(self,
                 time_columns = [],
                 list_of_feature = ['month',  'dayofweek', 'weekday', 'is_month_end', 'is_month_start', 'hour']
                 ):
        self.time_columns =time_columns
        self.list_of_feature = set(list_of_feature)
    def fit(self, X, y = None):
        LOGGER.info('FIT MakeTimeFeature')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if not self.time_columns:
            self.time_columns = X.select_dtypes(include = ['datetime64[ns]']).columns.tolist()
        return self
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM MakeTimeFeature')
        X = X.copy()
        if y is not None:
            y = y.copy()
        for col in self.time_columns:
            if 'hour' in self.list_of_feature:
                X[col+"_hour"] = X[col].dt.hour
            if 'dayofweek' in self.list_of_feature:
                X[col+"_dayofweek"] = X[col].dt.dayofweek
            if 'weekday' in self.list_of_feature:
                X[col + "_weekday"] = X[col].dt.weekday
            if 'is_month_start' in self.list_of_feature:
                X[col + "_is_month_start"] = X[col].dt.is_month_start
            if 'is_month_end' in self.list_of_feature:
                X[col + "_is_month_end"] = X[col].dt.is_month_end
        X.drop(columns = self.time_columns, errors='ignore', inplace = True)
        return X, y
    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X, y)


@PreprocessFactory.register('preprocess-outlier')
class Outlier(BaseEstimator, TransformerMixin):
    def __init__(self, methods = ['knn', 'iforest', 'pca'], contamination = 0.2, random_state = 42,  verbose = True):
        self.contamination = contamination
        self.random_state = random_state
        self.methods = methods
    def fit(self, X, y = None):
        LOGGER.info('FIT OUTLIER')
        X = X.copy()
        if y is not None:
            y = y.copy()
        self.outlier = []
        if "knn" in self.methods:
            knn = pyod.models.knn.KNN(contamination = self.contamination)
            self.outlier.append(knn)
        if 'iforest' in self.methods:
            iforest = pyod.models.iforest.IForest(contamination= self.contamination)
            self.outlier.append(iforest)

        if 'pca' in self.methods:
            pca = pyod.models.pca.PCA(contamination=self.contamination)
            self.outlier.append(pca)

    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM OUTLIER')
        X = X.copy()
        if y is not None:
            y = y.copy()
        X['vote_outlier'] = 0
        for out in self.outlier:
            X['vote_outlier'] += out.fit_predict(X.drop(columns=['vote_outlier']))
        LOGGER.warning('Remove outlier: {} rows'.format((X['vote_outlier']==len(self.methods)).sum()))
        if y is not None:
            y = y.loc[X['vote_outlier']!=len(self.methods)]
        X = X.loc[X['vote_outlier']!=len(self.methods)].drop(columns=['vote_outlier'])
        return X, y
    def fit_transform(self, X, y = None):
        self.fit(X, y)
        return self.transform(X, y)


@PreprocessFactory.register('preprocess-makenonlinearfeature')
class MakeNonLinearFeature(BaseEstimator, TransformerMixin):
    def __init__(self,
                polynomial_columns = [],
                degree = 2,
                interaction_only = False,
                include_bias = False,
                other_nonlinear_feature = ["sin", "cos", "tan"]
                 ):
        self.polynomial_columns = polynomial_columns
        self.degree = degree
        self.interaction_only = interaction_only
        self.include_bias = include_bias
        self.other_nonlinear_feature = other_nonlinear_feature
        self.poly = sklearn.preprocessing.PolynomialFeatures(degree=degree,
                                                             interaction_only=interaction_only,
                                                             include_bias=include_bias)
    def fit(self, X, y):
        X = X.copy()
        if y is not None:
            y = y.copy()
        LOGGER.info('FIT MakeNonLinearFeature')
        if len(self.polynomial_columns) == 0:
            self.polynomial_columns = X.select_dtypes(include=[np.number]).columns
        if self.polynomial_columns is None:
            self.polynomial_columns = []
        return self
    def transform(self, X, y):
        LOGGER.info('TRANSFORM MakeNonLinearFeature')
        X = X.copy()
        if y is not None:
            y = y.copy()

        if len(self.polynomial_columns) > 0:
            self.poly.fit(X[self.polynomial_columns])
            poly_feature = self.poly.get_feature_names(input_features=self.polynomial_columns)
            data = self.poly.transform(X[self.polynomial_columns])

            for idx, col in enumerate(poly_feature):
                X[col] = data[:, idx]

            if 'sin' in self.other_nonlinear_feature:
                for col in self.polynomial_columns:
                    X[col + "_sin"] = np.sin(X[col])
            if 'cos' in self.other_nonlinear_feature:
                for col in self.polynomial_columns:
                    X[col + "_cos"] = np.cos(X[col])
            if 'tan' in self.other_nonlinear_feature:
                for col in self.polynomial_columns:
                    X[col + "_tan"] = np.tan(X[col])



        return X, y




@PreprocessFactory.register('preprocess-reducecategorywithcount')
class ReduceCategoricalWithCount(BaseEstimator, TransformerMixin):
    def __init__(self,
                 categorical_columns=[]):
        self.categorical_columns = categorical_columns
    def fit(self, X, y = None):
        LOGGER.info('FIT REDUCE CATEGORICAL WITH COUNT')
        X = X.copy()
        if y is not None:
            y = y.copy()
        self.data_count = {}
        for col in self.categorical_columns:
            if col not in self.data_count.keys():
                self.data_count[col] = {}
            self.data_count[col] = X[col].value_counts().to_dict()
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM REDUCE CATEGORICAL WITH COUNT')
        X = X.copy()
        if y is not None:
            y = y.copy()
        for col in self.categorical_columns:
            X[col].replace(self.data_count[col].keys(), self.data_count[col].values(), inplace=True)
        return X, y
    def fit_transform(self, X, y = None):
        self.fit(X, y)
        return self.transform(X, y)

@PreprocessFactory.register('preprocess-rfe')
class RecursiveFeatureElimination(BaseEstimator, TransformerMixin):
    def __init__(self,
                 estimator = None,
                 step = 1,
                 min_features_to_select = 2,
                 cv = 2):
        if cv == 0:
            self.selector = sklearn.feature_selection.RFE(estimator=estimator, step = step, n_features_to_select=min_features_to_select)
        else:
            self.selector = sklearn.feature_selection.RFECV(estimator=estimator, step = step, min_features_to_select=min_features_to_select, cv = cv, n_jobs=-1)
    def fit(self, X, y = None):
        X = X.copy()
        if y is not None:
            y = y.copy()
        self.selector.fit(X, y)
        return self
    def transform(self, X, y = None):
        X = X.copy()
        if y is not None:
            y = y.copy()
        X = X.loc[:, self.selector.support_]
        return X, y
    def fit_transform(self, X, y = None):
        self.fit(X, y)
        return self.transform(X)
    def get_support(self, indices = False):
        return self.selector.get_support(indices=False)
    def get_ranking(self):
        return self.selector.ranking_

@PreprocessFactory.register('preprocess-reducedimension')
class ReduceDimension(BaseEstimator, TransformerMixin):
    def __init__(self,
                 method = 'pca_linear',
                 n_components = 0.99,
                 random_state = 42):
        self.method = method
        self.n_components = n_components
        self.random_state = random_state
    def fit(self, X, y = None):
        LOGGER.info('FIT REDUCE DIMENSION')
        X = X.copy()
        if y is not None:
            y = y.copy()
        if self.method == 'pca_linear':
            self.model = sklearn.decomposition.PCA(self.n_components, random_state=self.random_state).fit(X)
        if self.method == 'kernal_pca':
            self.model = sklearn.decomposition.KernelPCA(self.n_components, random_state=self.random_state).fit(X)
        if self.method == 'tsne':
            self.model = sklearn.manifold.TSNE(self.n_components, random_state=self.random_state).fit(X)
        if self.method == 'incremental':
            self.model = sklearn.decomposition.IncrementalPCA(self.n_components, random_state = random_state)
        return self
    def transform(self, X, y = None):
        LOGGER.info('TRANSFORM REDUCE DIMENSION')
        X = X.copy()
        if y is not None:
            y = y.copy()
        X = self.model.transform(X)
        return X, y
    def fit_transform(self, X, y = None):
        X = X.copy()
        self.fit(X, y)
        return self.transform(X, y)

def test(dataset, y):
    print(dataset.info())
    print('Start SimpleImputer ...')
    dataset = SimpleImputer(numeric_strategy='mean', categorical_strategy='most_frequent').fit_transform(dataset)
    print('Done SimpleImputer')
    print('Start Binning ...')
    dataset = Binning(features_to_discretize=['Jumps']).fit_transform(dataset)
    print('Done Binning')
    print('dataset: ', dataset)

    print('Start scaling ...')
    dataset = Scaling(numeric_columns=[], method = 'zscore').fit_transform(dataset)
    print('Done Binning')

    print('Start reducecategorywithcount ...')
    dataset = ReduceCategoricalWithCount(categorical_columns=['Jumps']).fit_transform(dataset)
    print('Done reducecategorywithcount')
    print(dataset)
    print('Start Outlier ...')
    dataset, y = Outlier().fit_transform(dataset, y)
    print('Done Outlier')

    kaka = RecursiveFeatureElimination(estimator=sklearn.ensemble.RandomForestClassifier(), min_features_to_select=3, cv=3).fit(dataset, y)
    print('selector: ',  kaka.get_support())
    print('Start RecursiveFeatureElimination ...')
    dataset = RecursiveFeatureElimination(estimator=sklearn.ensemble.RandomForestClassifier()).fit_transform(dataset, y)
    print('Done RecursiveFeatureElimination')
    print(dataset)
    print('Start ReduceDimensionForSupervised ...')
    dataset = ReduceDimensionForSupervised().fit_transform(dataset)
    print('Done ReduceDimensionForSupervised')
    print('dataset: ', dataset)
if __name__=='__main__':
    pass







from lightgbm import LGBMRegressor
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution

@ModelFactory.register('regression-lgbmregressor')
class LGBMRegressorContainer(LGBMRegressor):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        tune_grid = {
            "num_leaves": [10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 150, 200],
            "learning_rate": np_list_arange(0.001, 0.5, 0.001, inclusive=True),
            "n_estimators": np_list_arange(10, 300, 10, inclusive=True),
            "min_split_gain": [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
            "reg_alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "reg_lambda": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "feature_fraction": np_list_arange(0.4, 1, 0.1, inclusive=True),
            "bagging_fraction": np_list_arange(0.4, 1, 0.1, inclusive=True),
            "bagging_freq": [1, 2, 3, 4, 5, 6, 7],
            "min_child_samples": np_list_arange(5, 100, 5, inclusive=True),
        }
        tune_distributions = {
            "num_leaves": IntUniformDistribution(10, 200),
            "learning_rate": UniformDistribution(0.000001, 0.5, log=True),
            "n_estimators": IntUniformDistribution(10, 300),
            "min_split_gain": UniformDistribution(0, 1),
            "reg_alpha": UniformDistribution(0.0000000001, 10, log=True),
            "reg_lambda": UniformDistribution(0.0000000001, 10, log=True),
            "min_data_in_leaf": IntUniformDistribution(10, 10000),
            "feature_fraction": UniformDistribution(0.4, 1),
            "bagging_fraction": UniformDistribution(0.4, 1),
            "bagging_freq": IntUniformDistribution(1, 7),
            "min_child_samples": IntUniformDistribution(5, 100),
        }

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = LGBMRegressor(**kwargs)



import logging
from logging.handlers import RotatingFileHandler


class LogFilter(object):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno <= self.__level


def setup_logger():
    MAX_BYTES = 10000000  # Maximum size for a log file
    BACKUP_COUNT = 9  # Maximum number of old log files

    # The name should be unique, so you can get in in other places
    # by calling `logger = logging.getLogger('com.logger.example')
    logger = logging.getLogger('com.logger.example')
    logger.setLevel(logging.INFO)  # the level should be the lowest level set in handlers

    log_format = logging.Formatter('[%(levelname)s] %(asctime)s - %(message)s')

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(log_format)
    stream_handler.setLevel(logging.INFO)
    logger.addHandler(stream_handler)

    info_handler = RotatingFileHandler('info.log', maxBytes=MAX_BYTES, backupCount=BACKUP_COUNT)
    info_handler.setFormatter(log_format)
    info_handler.setLevel(logging.INFO)
    info_handler.addFilter(LogFilter(logging.INFO))
    logger.addHandler(info_handler)

    error_handler = RotatingFileHandler('error.log', maxBytes=MAX_BYTES, backupCount=BACKUP_COUNT)
    error_handler.setFormatter(log_format)
    error_handler.setLevel(logging.ERROR)
    logger.addHandler(error_handler)

    return logger


LOGGER = setup_logger()



from .adaboost_classifier import AdaBoostClassifierContainer
from .catboost_classifier import CatBoostClassifierContainer
from .decisiontree_classifier import DecisionTreeClassifierContainer
from .extratrees_classifier import ExtraTreesClassifierContainer
from .gaussianNB_classifier import GaussianNBContainer
from .gradient_boosting_classifier import GradientBoostingClassifierContainer
from .kneighbors_classifier import KNeighborsClassifierContainer
from .lgbm_classifier import LGBMClassifierContainer
from .lineardiscriminantanalysis import LinearDiscriminantAnalysisContainer
from .logisticregression import LogisticRegressionContainer
from .mlpclassifier import MLPClassifierContainer
from .quadraticdiscriminantanalysis import QuadraticDiscriminantAnalysisContainer
from .randomforest_classifier import RandomForestClassifierContainer
# from .ridge_classifier import RidgeClassifierContainer
# from .sgd_classifier import SGDClassifierContainer
# from .svc_classifier import SVCClassifierContainer
from .xgb_classifier import XGBClassifierContainer




from sklearn.neural_network import MLPClassifier
from autotonne.models.model_factory import ModelFactory
from autotonne.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution, CategoricalDistribution
@ModelFactory.register('classification-mlpclassifier')
class MLPClassifierContainer():
    def __init__(self, **kwargs):
        tune_grid = {}
        tune_distributions = {}
        tune_grid = {
            "learning_rate": ["constant", "invscaling", "adaptive"],
            "alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                0.9,
            ],
            "activation": ["tanh", "identity", "logistic", "relu"],
        }
        tune_distributions = {
            "alpha": UniformDistribution(0.0000000001, 0.9999999999, log=True),
            "activation": CategoricalDistribution(["tanh", "identity", "logistic", "relu"]),
            "learning_rate": CategoricalDistribution(["constant", "invscaling", "adaptive"])
        }
        self.tune_grid = tune_grid

        self.tune_distributions = tune_distributions
        self.estimator = MLPClassifier(**kwargs)




from abc import ABCMeta, abstractmethod

class ModelFactory(object):
    registry = {}
    name_registry = []
    @classmethod
    def register(cls, name: str):
        def inner_wrapper(wrapped_class):
            if name in cls.registry:
                print('Executor {} already axists. Will replace it'.format(name))
            cls.registry[name] = wrapped_class
            cls.name_registry.append(name)
            return wrapped_class
        return inner_wrapper
    @classmethod
    def create_executor(cls, name, **kwargs):
        if name not in cls.registry:
            print('Executor {} does not exist in the registry'.format(name))
            return None
        exec_class = cls.registry[name]
        executor = exec_class(**kwargs)
        return executor



from .logger import LOGGER
from .utils import can_early_stop, save_model, load_model
from .distributions import np_list_arange, UniformDistribution, IntUniformDistribution
from .cloud_service import CloudService
